@@ -23,6 +23,10 @@ define('INSTALL_PATH', realpath(__DIR__ . '/..') . '/' );
 
 require_once INSTALL_PATH . 'program/include/clisetup.php';
 
+if (!function_exists('system')) {
+  rcube::raise_error("PHP system() function is required. Check disable_functions in php.ini.", false, true);
+}
+
 $target_dir = unslashify($_SERVER['argv'][1]);
 
 if (empty($target_dir) || !is_dir(realpath($target_dir)))
@@ -50,19 +54,19 @@ if (strtolower($input) == 'y') {
   }
 
   $dirs = array('program','installer','bin','SQL','plugins','skins');
-  if (is_dir(INSTALL_PATH . 'vendor') && !is_file(INSTALL_PATH . 'composer.json')) {
+  if (is_dir(INSTALL_PATH . 'vendor') && !is_file("$target_dir/composer.json")) {
     $dirs[] = 'vendor';
   }
   foreach ($dirs as $dir) {
     // @FIXME: should we use --delete for all directories?
-    $delete  = in_array($dir, array('program', 'installer')) ? '--delete ' : '';
-    $command = "rsync -aC --out-format \"%n\" " . $delete . INSTALL_PATH . "$dir/* $target_dir/$dir/";
+    $delete  = in_array($dir, array('program', 'installer', 'vendor')) ? '--delete ' : '';
+    $command = "rsync -aC --out-format=%n " . $delete . INSTALL_PATH . "$dir/ $target_dir/$dir/";
     if (!system($command, $ret) || $ret > 0) {
       rcube::raise_error("Failed to execute command: $command", false, true);
     }
   }
-  foreach (array('index.php','.htaccess','config/defaults.inc.php','composer.json-dist','CHANGELOG','README.md','UPGRADING','LICENSE','INSTALL') as $file) {
-    $command = "rsync -a --out-format \"%n\" " . INSTALL_PATH . "$file $target_dir/$file";
+  foreach (array('index.php','.htaccess','config/defaults.inc.php','composer.json-dist','jsdeps.json','CHANGELOG','README.md','UPGRADING','LICENSE','INSTALL') as $file) {
+    $command = "rsync -a --out-format=%n " . INSTALL_PATH . "$file $target_dir/$file";
     if (file_exists(INSTALL_PATH . $file) && (!system($command, $ret) || $ret > 0)) {
       rcube::raise_error("Failed to execute command: $command", false, true);
     }
@@ -95,6 +99,21 @@ if (strtolower($input) == 'y') {
       echo "done.\n\n";
   }
 
+  // check if js-deps are up-to-date
+  if (file_exists("$target_dir/jsdeps.json") && file_exists("$target_dir/bin/install-jsdeps.sh")) {
+    $jsdeps = json_decode(file_get_contents("$target_dir/jsdeps.json"));
+    $package = $jsdeps->dependencies[0];
+    $dest_file = $target_dir . '/' . $package->dest;
+    if (!file_exists($dest_file) || sha1_file($dest_file) !== $package->sha1) {
+        echo "Installing JavaScript dependencies...";
+        system("cd $target_dir && bin/install-jsdeps.sh");
+        echo "done.\n\n";
+    }
+  }
+  else {
+    echo "JavaScript dependencies installation skipped...\n";
+  }
+
   echo "Running update script at target...\n";
   system("cd $target_dir && php bin/update.sh --version=$oldversion");
   echo "All done.\n";
